-- NOTE: THIS SCRIPT MUST BE RUN IN SQLCMD MODE TO FUNCTION PROPERLY
:CONNECT Buzz

-- Switch to the InitiatorDB database
USE InitiatorDB;
GO

CREATE MESSAGE TYPE [EndofStream]
VALIDATION = NONE;

-- 
-- This table associates the current connection (@@SPID)
-- with a conversation. The association key also
-- contains the conversation parameteres:
-- from service, to service, contract used
--
CREATE TABLE [SessionConversations] (
	SPID INT NOT NULL	
	, FromService SYSNAME NOT NULL
	, ToService SYSNAME NOT NULL
	, ONContract SYSNAME NOT NULL
	, Handle UNIQUEIDENTIFIER NOT NULL
	, PRIMARY KEY (SPID, FromService, ToService, ONContract)
	, UNIQUE (Handle));
GO

-- SEND procedure. Will lookup to reuse an existing conversation
-- or start a new in case no conversation exists or the conversation
-- cannot be used
--
CREATE PROCEDURE [usp_Send] (
@fromService SYSNAME,
@toService SYSNAME,
@onContract SYSNAME,
@messageType SYSNAME,
@messageBody VARBINARY(MAX))
AS
BEGIN
	SET NOCOUNT ON;
	DECLARE @handle UNIQUEIDENTIFIER;
	DECLARE @counter INT;
	DECLARE @error INT;
	SELECT @counter = 1;
	BEGIN TRANSACTION;
	-- Will need a loop to retry in case the conversation is
	-- in a state that does not allow transmission
	--
	WHILE (1=1)
	BEGIN
		-- Seek an eligible conversation in [SessionConversations]
		--
		SELECT @handle = Handle
			FROM [SessionConversations] WITH (UPDLOCK)
			WHERE SPID = @@SPID
			AND FromService = @fromService
			AND ToService = @toService
			AND ONContract = @OnContract;
		IF @handle IS NULL
		BEGIN
			-- Need to start a new conversation for the current @@spid
			--
			BEGIN DIALOG CONVERSATION @handle
				FROM SERVICE @fromService
				TO SERVICE @toService
				ON CONTRACT @onContract
				WITH ENCRYPTION = ON;
			-- Set an ONe minute timer ON the conversation
			--
			BEGIN CONVERSATION TIMER (@handle) TIMEOUT = 60;
			INSERT INTO [SessionConversations]
				(SPID, FromService, ToService, ONContract, Handle)
				VALUES
				(@@SPID, @fromService, @toService, @onContract, @handle);
		END;
		-- Attempt to SEND ON the associated conversation
		--
		SEND ON CONVERSATION @handle
			MESSAGE TYPE @messageType
			(@messageBody);
		SELECT @error = @@ERROR;
		IF @error = 0
		BEGIN
			-- Successful send, just exit the loop
			--
			BREAK;
		END
		SELECT @counter = @counter+1;
		IF @counter > 10
		BEGIN
			-- We failed 10 times in a row, something must be broken
			--
			RAISERROR (
				N'Failed to SEND ON a conversation for more than 10 times. Error %i.'
				, 16, 1, @error) WITH LOG;
			BREAK;
		END
		-- Delete the associated conversation from the table and try again
		--
		DELETE FROM [SessionConversations]
			WHERE Handle = @handle;
		SELECT @handle = NULL;
	END
	COMMIT;
END
GO

-- NOTE: THIS SCRIPT MUST BE RUN IN SQLCMD MODE TO FUNCTION PROPERLY
:CONNECT Buzz
USE INITIATORDB;
GO
ALTER PROCEDURE InitiatorQueue_ActivationProcedure
AS
BEGIN
	DECLARE @handle UNIQUEIDENTIFIER;
	DECLARE @messageTypeName SYSNAME;
	DECLARE @messageBody VARBINARY(MAX);
	BEGIN TRANSACTION;
	RECEIVE TOP(1)
		@handle = conversation_handle,
		@messageTypeName = message_type_name,
		@messageBody = message_body
		FROM InitiatorQueue;
	IF @handle IS NOT NULL
	BEGIN
		-- Delete the message from the [SessionConversations] table
		-- before sending the [EndOfStream] message. The order is
		-- important to avoid deadlocks. Strictly speaking, we should
		-- ONly delete if the message type is timer or error, but is
		-- simpler and safer to just delete always
		--

		IF @messageTypeName = N'http://schemas.microsoft.com/SQL/ServiceBroker/DialogTimer'
		BEGIN
			DELETE FROM [SessionConversations]
			WHERE [Handle] = @handle;

			SEND ON CONVERSATION @handle 
				MESSAGE TYPE [EndOfStream];
		END
		ELSE IF @messageTypeName = N'http://schemas.microsoft.com/SQL/ServiceBroker/EndDialog'
		BEGIN
			END CONVERSATION @handle;
		END
		ELSE IF @messageTypeName = N'http://schemas.microsoft.com/SQL/ServiceBroker/Error'
		BEGIN
			END CONVERSATION @handle;
			-- Insert your error handling here. Could send a notification or
			-- store the error in a table for further inspection
			-- Were gonna log the error into the ERRORLOG and Eventvwr.exe
			--
			DECLARE @error INT;
			DECLARE @description NVARCHAR(4000);
			WITH XMLNAMESPACES ('http://schemas.microsoft.com/SQL/ServiceBroker/Error' AS ssb)
			SELECT
				@error = CAST(@messageBody AS XML).value(
					'(//ssb:Error/ssb:Code)[1]', 'INT'),
				@description = CAST(@messageBody AS XML).value(
					'(//ssb:Error/ssb:Description)[1]', 'NVARCHAR(4000)')
			RAISERROR(N'Received error Code:%i Description:"%s"',
				 16, 1, @error, @description) WITH LOG;
		END;
	END
	COMMIT;
END
GO

-- attach the uspSenderActivation procedure with the senders queue
--
ALTER QUEUE [InitiatorQueue]
	WITH ACTIVATION (
	STATUS = ON,
	MAX_QUEUE_READERS = 10,
	PROCEDURE_NAME = [InitiatorQueue_ActivationProcedure],
	EXECUTE AS OWNER);
GO